/********************************************************************************
    LibVideoCoding - video coding library
    Copyright (C) 2002  Dirk Farin

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ********************************************************************************/

#include "libvideocoding/mpeg2/syntax/system.hh"
#include "libvideocoding/mpeg2/syntax/errors.hh"
#include <libvideogfx/utility/bitstream/bitreader.hh>

#include <iostream>
using namespace std;

namespace videocoding {
  using namespace videogfx;

  // ByteBufferPool MP2_SystemPacket::s_bytebufferpool;


  void MP2_PackHeader::Read(InputStream& str)
  {
    ByteBuffer buf;
    ReadUntilNextStartCode(buf,str);   // read the complete pack-header

    BitReader reader(buf.AskData(), buf.AskLength());

    // check start-code and first two ID-bits

    uint32 sc = reader.GetBits(32);
    if (sc != (0x100 | MP2_StartCode_PackStart))
      throw Excpt_InvalidData("pack header does not start with pack start code");

    int id = reader.PeekBits(4);
    bool MPEG1 = (id == 2);

    // make sure that enough bytes are contained in the pack-header

    int header_length = (MPEG1 ? 12 : 14);

    if (buf.AskLength() < header_length)
      throw Excpt_UnexpectedEndOfData("pack header");

    if (MPEG1)
      {
	reader.SkipBits(4);
      }
    else
      {
	if (reader.GetBits(2) != 1)
	  throw Excpt_InvalidData("MPEG-2 marker in pack header is not '01'");
      }


    // read pack-header

    SCR = ((uint64)reader.GetBits(3)) << 30;
    reader.SkipBits(1);
    SCR |= reader.GetBits(15) << 15;
    reader.SkipBits(1);
    SCR |= reader.GetBits(15);
    reader.SkipBits(1);

    SCR *= 300;

    if (!MPEG1) SCR += reader.GetBits(9);

    reader.SkipBits(1);
    mux_rate = reader.GetBits(22);
    reader.SkipBits(1);

    if (MPEG1)
      {
	n_stuffing_bytes = 0;
      }
    else
      {
	reader.SkipBits(1+5);
	n_stuffing_bytes = reader.GetBits(3);

	// handle stuffing bytes

	header_length += n_stuffing_bytes;

	if (buf.AskLength() < header_length)
	  throw Excpt_UnexpectedEndOfData("pack header");

	/*
	  for (int i=0;i<n_stuffing_bytes;i++)
	  reader.SkipBits(8);
	*/
      }
  }


  void MP2_PackHeader::Show(ostream& ostr) const
  {
    ostr << "SCR:            " << SCR << endl
	 << "mux rate:       " << mux_rate << endl
	 << "stuffing bytes: " << n_stuffing_bytes << endl;

  }


  MP2_SystemHeader::MP2_SystemHeader()
  {
    for (int i=0xBC ; i<0x100 ; i++)
      STD_buffer_size_bound[i-0xBC] = 0;
  }


  void  MP2_SystemHeader::Read(InputStream& str)
  {
    ByteBuffer buf;
    ReadUntilNextStartCode(buf,str);   // read the complete header

    BitReader reader(buf.AskData(), buf.AskLength());
  
    // check start-code

    uint32 sc = reader.GetBits(32);
    if (sc != (0x100 | MP2_StartCode_SystemHeaderStart))
      throw Excpt_InvalidData("system header does not start with system-header start code");


    // make sure that enough bytes are contained in the pack-header

    int header_length = reader.GetBits(16) + 4+2;
    if (buf.AskLength() < header_length)
      throw Excpt_UnexpectedEndOfData("system header");

    reader.SkipBits(1);
    rate_bound = reader.GetBits(22);
    reader.SkipBits(1);
    audio_bound = reader.GetBits(6);
    fixed_bitrate = reader.GetBits(1);
    CSPS_flag     = reader.GetBits(1);
    audio_lock    = reader.GetBits(1);
    video_lock    = reader.GetBits(1);
    reader.SkipBits(1);
    video_bound   = reader.GetBits(5);
    packet_rate_restriction_flag = reader.GetBits(1);
    reader.SkipBits(7);

    while (reader.PeekBits(1)==1 && reader.AskBitsLeft()>0)
      {
	uint8 stream_id = reader.GetBits(8);
	reader.SkipBits(2);
	int scale = reader.GetBits(1);
	int bound = reader.GetBits(13);
	if (scale==0)
	  bound *= 128;
	else
	  bound *= 1024;

	if (stream_id < 0xBC && stream_id != 0xB8 && stream_id != 0xB9)
	  throw Excpt_InvalidData("P-STD buffer size specified for undefined stream");

	if (stream_id == 0xB8)
	  for (int i=0xC0; i<0xE0; i++)
	    STD_buffer_size_bound[i-0xBC] = bound;
	else if (stream_id == 0xB9)
	  for (int i=0xE0; i<0xF0; i++)
	    STD_buffer_size_bound[i-0xBC] = bound;
	else
	  STD_buffer_size_bound[stream_id-0xBC] = bound;
      }
  }


  void MP2_SystemHeader::Show(ostream& ostr) const
  {
    ostr << "rate bound:        " << rate_bound << endl
	 << "audio bound:       " << ((int)audio_bound) << endl
	 << "video bound:       " << ((int)video_bound) << endl
	 << "fixed bitrate:     " << (fixed_bitrate ? "yes":"no") << endl
	 << "CSPS:              " << (CSPS_flag ? "set" : "clear") << endl
	 << "audio lock:        " << (audio_lock ? "on" : "off") << endl
	 << "video lock:        " << (video_lock ? "on" : "off") << endl
	 << "packet rate restr.:" << (packet_rate_restriction_flag ? "set":"clear") << endl;

    for (int i=0xBC ; i<0x100; i++)
      {
	if (STD_buffer_size_bound[i-0xBC])
	  {
	    if (IsMPEG_VIDEO_STREAM((MP2_StartCode)i)) ostr << "Video ";
	    else if (IsMPEG_AUDIO_STREAM((MP2_StartCode)i)) ostr << "Audio ";
	    else ostr << "Misc  ";

	    ostr << "(" << std::hex << ((int)i) << std::dec << ") :       ";

	    ostr << STD_buffer_size_bound[i-0xBC] << endl;
	  }
      }
  }



  void  MP2_PES_Packet::Reset()
  {
    stream_id = 0;

    PES_scrambling_control = 0;
    PES_priority = false;
    data_alignment_indicator = false;
    copyright = false;
    original  = true;

    has_PTS = false;
    has_DTS = false;
    has_ESCR = false;
    has_ES_rate = false;
    has_DSM_trick_mode = false;
    has_additional_copy_info = false;
    has_PES_CRC = false;
    has_PES_extension = false;
    has_PES_private_data = false;
    has_pack_header_field = false;
    has_program_packet_sequence_counter = false;
    has_PSTD_buffer = false;
    has_PES_extension_flag_2 = false;
  }

  void  MP2_PES_Packet::Read(InputStream& istr)
  {
    Reset();

    uint8 buf[6];

    int size = istr.FillBuffer(buf,6);
    if (size<6)
      throw Excpt_UnexpectedEndOfData("PES header");

    if (buf[0]!=0 || buf[1]!=0 || buf[2]!=1)
      throw Excpt_InvalidData("PES packet does not begin with start code");

    stream_id = buf[3];

    int packet_len = (buf[4]<<8) | buf[5];

    uint8* dataptr;

    bool length_specified = (packet_len != 0);

    if (length_specified)
      {
	dataptr = data.GetPtrToAppendToBuffer(packet_len);
	size = istr.FillBuffer(dataptr,packet_len,packet_len);
	if (size != packet_len)
	  throw Excpt_UnexpectedEndOfData("PES packet data");
      }
    else
      {
	dataptr = data.GetPtrToAppendToBuffer(3);
	size = istr.FillBuffer(dataptr,3,3);
	if (size != 3)
	  throw Excpt_UnexpectedEndOfData("PES packet data");

	int headerlen = dataptr[2];
	uint8* moreheader = data.GetPtrToAppendToBuffer(headerlen);
	size = istr.FillBuffer(moreheader,headerlen,headerlen);
	if (size != headerlen)
	  throw Excpt_UnexpectedEndOfData("PES packet data");
      }

    int hdrlen = 0;
    int header_length = 0;  // length as specified in stream  (==0 -> no header length specified)

    if (stream_id != MP2_StartCode_Program_Stream_Map &&
	stream_id != MP2_StartCode_Padding_Stream     &&
	stream_id != MP2_StartCode_Private_Stream_2   &&
	stream_id != MP2_StartCode_ECM_Stream         &&
	stream_id != MP2_StartCode_EMM_Stream         &&
	stream_id != MP2_StartCode_Program_Stream_Directory &&
	stream_id != MP2_StartCode_DSMCC_Stream       &&
	stream_id != MP2_StartCode_H222_Type_E)
      {
	BitReader reader(dataptr, packet_len);

	while (reader.PeekBits(8)==0xFF)
	  {
	    reader.SkipBits(8);
	    hdrlen++;
	  }

	bool ISO11172 = true;

	if (reader.PeekBits(2)==1)
	  {
	    reader.SkipBits(2);

	    has_PSTD_buffer = true;
	    int scale = (reader.GetBits(1) ? 1024:128);
	    STDBufferSize = reader.GetBits(13)*scale;

	    hdrlen += 2;
	  }
	else if (reader.PeekBits(2)==2)
	  {
	    reader.SkipBits(2);
	    ISO11172 = false;

	    PES_scrambling_control = reader.GetBits(2);
	    PES_priority = reader.GetBits(1);
	    data_alignment_indicator = reader.GetBits(1);
	    copyright = reader.GetBits(1);
	    original  = reader.GetBits(1);

	    has_PTS = reader.GetBits(1);
	    has_DTS = reader.GetBits(1);
	    if (has_DTS && !has_PTS) throw Excpt_InvalidData("PES header has DTS but no PTS");
	    has_ESCR = reader.GetBits(1);
	    has_ES_rate = reader.GetBits(1);
	    has_DSM_trick_mode = reader.GetBits(1);

	    has_additional_copy_info = reader.GetBits(1);
	    has_PES_CRC = reader.GetBits(1);
	    has_PES_extension = reader.GetBits(1);
	    header_length = reader.GetBits(8);

	    if (header_length) header_length += 3;

	    hdrlen += 3;
	  }

	uint4 PTS_DTS_id = reader.PeekBits(4);
	if (has_PTS || (ISO11172 &&(PTS_DTS_id & 0xE)==0x2))
	  {
	    reader.SkipBits(4);
	    has_PTS = true;

	    PTS = ((uint64)reader.GetBits(3))<<30;
	    reader.SkipBits(1);
	    PTS |= reader.GetBits(15)<<15;
	    reader.SkipBits(1);
	    PTS |= reader.GetBits(15);
	    reader.SkipBits(1);

	    hdrlen += 5;
	  }

	if (has_DTS || (ISO11172 && PTS_DTS_id == 0x3))
	  {
	    uint4 id = reader.GetBits(4);
	    if (id != 0x1) throw Excpt_InvalidData("system header: marker between PTS & DTS is not '0001'");

	    has_DTS = true;

	    DTS = ((uint64)reader.GetBits(3))<<30;
	    reader.SkipBits(1);
	    DTS |= reader.GetBits(15)<<15;
	    reader.SkipBits(1);
	    DTS |= reader.GetBits(15);
	    reader.SkipBits(1);

	    hdrlen += 5;
	  }

	if (PTS_DTS_id == 0 && ISO11172)
	  {
	    uint8 pattern = reader.GetBits(8);
	    if (pattern != 0x0F)
	      throw Excpt_InvalidData("ISO 11172 system header: private stream 2 pattern is not '0000 1111'");

	    hdrlen++;
	  }

	if (has_ESCR)
	  {
	    reader.SkipBits(2);
	    ESCR = ((uint64)reader.GetBits(3))<<30;
	    reader.SkipBits(1);
	    ESCR |= reader.GetBits(15)<<15;
	    reader.SkipBits(1);
	    ESCR |= reader.GetBits(15);
	    reader.SkipBits(1);
	    ESCR *= 300;
	    ESCR += reader.GetBits(9);
	    reader.SkipBits(1);

	    hdrlen += 6;
	  }

	if (has_ES_rate)
	  {
	    reader.SkipBits(1);
	    ES_rate = reader.GetBits(22);
	    reader.SkipBits(1);

	    hdrlen += 3;
	  }

	if (has_DSM_trick_mode)
	  {
	    trick_mode_control = (MP2_TrickModeControl)reader.GetBits(3);

	    switch (trick_mode_control)
	      {
	      MP2_TrickMode_FastForward:
	      MP2_TrickMode_FastReverse:
		field_id = (MP2_FieldControl)reader.GetBits(2);
		intra_slice_refresh = reader.GetBits(1);
		frequency_truncation = (MP2_CoefficientSelection)reader.GetBits(2);
		break;
	      MP2_TrickMode_SlowMotion:
	      MP2_TrickMode_SlowReverse:
		rep_cntrl = reader.GetBits(5);
		break;
	      MP2_TrickMode_FreezeFrame:
		reader.SkipBits(5);
		break;
	      }

	    hdrlen++;
	  }

	if (has_additional_copy_info)
	  {
	    reader.SkipBits(1);
	    additional_copy_info = reader.GetBits(7);

	    hdrlen++;
	  }

	if (has_PES_CRC)
	  {
	    previous_PES_packet_CRC = reader.GetBits(16);

	    hdrlen += 2;
	  }

	if (has_PES_extension)
	  {
	    has_PES_private_data  = reader.GetBits(1);
	    has_pack_header_field = reader.GetBits(1);
	    has_program_packet_sequence_counter = reader.GetBits(1);
	    has_PSTD_buffer = reader.GetBits(1);
	    reader.SkipBits(3);
	    has_PES_extension_flag_2 = reader.GetBits(1);

	    hdrlen++;

	    if (has_PES_private_data)
	      {
		for (int i=0;i<16;i++)
		  PES_private_data[i] = reader.GetBits(8);

		hdrlen += 128;
	      }

	    if (has_pack_header_field)
	      {
		uint8 len = reader.GetBits(8);
		for (int i=0;i<len;i++) reader.SkipBits(len);  // TODO

		hdrlen += len+1;
	      }

	    if (has_program_packet_sequence_counter)
	      {
		reader.SkipBits(1);
		program_packet_sequence_counter = reader.GetBits(7);
		reader.SkipBits(1);
		MPEG1_MPEG2_identifier = reader.GetBits(1);
		original_stuff_length = reader.GetBits(6);

		hdrlen += 2;
	      }

	    if (has_PSTD_buffer)
	      {
		reader.SkipBits(2);

		has_PSTD_buffer = true;
		int scale = (reader.GetBits(1) ? 1024:128);
		STDBufferSize = reader.GetBits(13)*scale;

		hdrlen += 2;
	      }

	    if (has_PES_extension_flag_2)
	      {
		reader.SkipBits(1);
		uint7 len = reader.GetBits(7);
		for (int i=0;i<len;i++)
		  reader.SkipBits(8);

		hdrlen += 1+len;
	      }
	  }

	int hl = packet_len - reader.AskBitsLeft()/8;
	Assert(hdrlen == hl);

	if (header_length && hdrlen > header_length)
	  throw Excpt_InvalidData("system header: header was longer than specified length");

	if (header_length)
	  {
	    while (header_length > hdrlen)
	      {
		reader.SkipBits(8);
		hdrlen++;
	      }
	  }
      }

    // data_start_offset = ; // number of bytes between 'absolute_filepos' and this data

    // remove header-data from byte-buffer

    data_start_offset = hdrlen+6;

    if (length_specified)
      {
	for (int i=hdrlen;i<packet_len;i++)
	  dataptr[i-hdrlen] = dataptr[i];

	data.TruncateBuffer(hdrlen);
      }
    else
      {
	// video PES-header with packet_length==0

	data.Clear();

	for (;;)
	  {
	    ReadUntilNextStartCode(data, istr);

	    uint8 buf[4];
	    if (istr.FillBuffer(buf,4,4) != 4)
	      break;

	    Assert(buf[0]==0 && buf[1]==0 && buf[2]==1);
	    if (IsMPEG_PACKETCODE((MP2_StartCode)buf[3]))
	      {
		istr.Pushback(buf,4);
		break;
	      }
	    else
	      {
		/* We have to push back the start code.
		   Copying to the output and continuing to read does not work because
		   this fails when the start code is at the very end of the PES packet.
		   The next call to ReadUntilNextStartCode() would overread the PES-packet
		   start code, since it skips start codes at the very beginning of the data. */

		istr.Pushback(buf,4);
	      }
	  }
      }
  }


  void  MP2_PES_Packet::Show(ostream& ostr) const
  {
    ostr << "data-offset: " << hex << data_start_offset << dec << endl;
    ostr << "stream id:              ";

    if (IsMPEG_VIDEO_STREAM((MP2_StartCode)stream_id))
      ostr << "video" << ((int)(stream_id-0xE0));
    else if (IsMPEG_AUDIO_STREAM((MP2_StartCode)stream_id))
      ostr << "audio" << ((int)(stream_id-0xC0));
    else
      switch (stream_id)
	{
	case MP2_StartCode_Program_Stream_Map: ostr << "program stream map"; break;
	case MP2_StartCode_Private_Stream_1:   ostr << "private stream 1"; break;
	case MP2_StartCode_Padding_Stream:     ostr << "padding stream"; break;
	case MP2_StartCode_Private_Stream_2:   ostr << "private stream 2"; break;
	case MP2_StartCode_ECM_Stream:         ostr << "ECM"; break;
	case MP2_StartCode_EMM_Stream:         ostr << "EMM"; break;
	case MP2_StartCode_DSMCC_Stream:       ostr << "DSMCC"; break;
	case MP2_StartCode_IEC13522_Stream:    ostr << "IEC13522"; break;
	case MP2_StartCode_H222_Type_A:        ostr << "H222 type A"; break;
	case MP2_StartCode_H222_Type_B:        ostr << "H222 type B"; break;
	case MP2_StartCode_H222_Type_C:        ostr << "H222 type C"; break;
	case MP2_StartCode_H222_Type_D:        ostr << "H222 type D"; break;
	case MP2_StartCode_H222_Type_E:        ostr << "H222 type E"; break;
	case MP2_StartCode_Ancillary_Stream:   ostr << "ancillary stream"; break;
	case MP2_StartCode_Program_Stream_Directory: ostr << "program stream directory"; break;
	}

    ostr << hex << " (" << ((int)stream_id) << ")" << dec << endl
	 << "PES scrambling control: " << ((int)PES_scrambling_control) << endl
	 << "PES priority:           " << (PES_priority ? "yes":"no") << endl
	 << "data alignment:         " << (data_alignment_indicator ? "yes":"no") << endl
	 << "copyright:              " << (copyright ? "yes":"no") << endl
	 << "original:               " << (original ? "yes":"no") << endl
	 << "PTS:                    ";
    if (has_PTS) ostr << PTS << endl; else ostr << "not present\n";
    ostr << "DTS:                    ";
    if (has_DTS) ostr << DTS << endl; else ostr << "not present\n";
    ostr << "ESCR:                   ";
    if (has_ESCR) ostr << ESCR << endl; else ostr << "not present\n";
    ostr << "ES rate:                ";
    if (has_ES_rate) ostr << ES_rate << endl; else ostr << "not present\n";

    ostr << "DSM trick mode:         ";
    if (!has_DSM_trick_mode)
      ostr << "not present\n";
    else
      {
	ostr << "trick mode:             ";
	switch (trick_mode_control)
	  {
	  case MP2_TrickMode_FastForward: ostr << "fast forward\n"; break;
	  case MP2_TrickMode_SlowMotion:  ostr << "slow motion\n"; break;
	  case MP2_TrickMode_FreezeFrame: ostr << "freeze frame\n"; break;
	  case MP2_TrickMode_FastReverse: ostr << "fast reverse\n"; break;
	  case MP2_TrickMode_SlowReverse: ostr << "slow reverse\n"; break;
	  }

	if (trick_mode_control == MP2_TrickMode_FastForward ||
	    trick_mode_control == MP2_TrickMode_FastReverse ||
	    trick_mode_control == MP2_TrickMode_FreezeFrame)
	  {
	    ostr << "field control:          ";
	    switch (field_id)
	      {
	      case MP2_FieldControl_TopFieldOnly:    ostr << "top field only\n"; break;
	      case MP2_FieldControl_BottomFieldOnly: ostr << "bottom field only\n"; break;
	      case MP2_FieldControl_CompleteFrame:   ostr << "complete frame\n"; break;
	      }
	  }

	if (trick_mode_control == MP2_TrickMode_FastForward ||
	    trick_mode_control == MP2_TrickMode_FastReverse)
	  {
	    ostr << "intra slice refresh:    " << (intra_slice_refresh ? "yes":"no") << endl;

	    ostr << "frequency truncation:   ";
	    switch (frequency_truncation)
	      {
	      case MP2_CoefficientSelection_OnlyDC:     ostr << "only DC\n"; break;
	      case MP2_CoefficientSelection_FirstThree: ostr << "first three coefficients\n"; break;
	      case MP2_CoefficientSelection_FirstSix:   ostr << "first six coefficients\n"; break;
	      case MP2_CoefficientSelection_All:        ostr << "all coefficients\n"; break;
	      }
	  }

	if (trick_mode_control == MP2_TrickMode_SlowMotion ||
	    trick_mode_control == MP2_TrickMode_SlowReverse)
	  {
	    ostr << "repetition count:       " << ((int)rep_cntrl) << endl;
	  }
      }

    ostr << "additional copy info:   " << (has_additional_copy_info ? "yes":"no") << endl;
    ostr << "has PES CRC:            " << (has_PES_CRC ? "yes":"no") << endl;

    ostr << "has pack header:        " << (has_pack_header_field ? "yes":"no") << endl;
    ostr << "program sequence counter: ";
    if (!has_program_packet_sequence_counter) ostr << "not present\n";
    else
      {
	ostr << ((int)program_packet_sequence_counter) << endl;
	ostr << "MPEG1 MPEG2 identifier: " << (MPEG1_MPEG2_identifier ? "yes":"no") << endl;
	ostr << "original stuffing:      " << ((int)original_stuff_length) << " bytes\n";
      }

    ostr << "P-STD buffer size:      ";
    if (has_PSTD_buffer)
      ostr << STDBufferSize << endl;
    else
      ostr << "not present\n";

    ostr << "payload length: " << data.AskLength() << " bytes\n";

#if 0
    ostr << hex
	 << ((int)data.AskData()[0]) << " "
	 << ((int)data.AskData()[1]) << " "
	 << ((int)data.AskData()[2]) << " "
	 << ((int)data.AskData()[3]) << " ... "
	 << ((int)data.AskData()[data.AskLength()-4]) << " "
	 << ((int)data.AskData()[data.AskLength()-3]) << " "
	 << ((int)data.AskData()[data.AskLength()-2]) << " "
	 << ((int)data.AskData()[data.AskLength()-1]) << dec << endl;
#endif
  }



  void  MP2_Transport_Packet::Read(InputStream& istr)
  {
    input_filepos = istr.AskPosition();

    uint8 buf[4];
    int len = istr.FillBuffer(buf,4,4);
    if (len<4)
      throw Excpt_UnexpectedEndOfData("transport header");

    BitReader bs(buf,4);

    uint4 sync_byte = bs.GetBits(8);
    if (sync_byte != 0x47)
      throw Excpt_InvalidData("transport packet does not start with sync byte 0x47");

    transport_error_indicator = bs.GetBits(1);
    payload_unit_start_indicator = bs.GetBits(1);
    transport_priority = bs.GetBits(1);
    PID = bs.GetBits(13);
    transport_scrambling_control = bs.GetBits(2);
    has_adaptation_field = bs.GetBits(1);
    has_payload = bs.GetBits(1);
    if (!has_payload && !has_adaptation_field)
      throw Excpt_InvalidData("transport packet has no adaptiation field and no payload");
    continuity_counter = bs.GetBits(4);

    payload_length=184;

    if (has_adaptation_field)
      {
	// TODO: decoding of adaption field

	uint8 length;
	if (istr.FillBuffer(&length,1,1) != 1)
	  throw Excpt_UnexpectedEndOfData("transport header");

	payload_length -= length+1;

	// skip adaptation field by reading it into the payload buffer (this is only a fake)
	if (length != istr.FillBuffer(payload,length,length))
	  throw Excpt_UnexpectedEndOfData("transport header adaptation field");
      }

    if (payload_length != (int)istr.FillBuffer(payload,payload_length,payload_length))
      throw Excpt_UnexpectedEndOfData("transport header payload");

    if (!has_payload)
      payload_length = 0;
  }


  void  MP2_Transport_Packet::Show(ostream& ostr) const
  {
    ostr << "transport error:    " << (transport_error_indicator ? "yes" : "no") << endl
	 << "payload unit start: " << (payload_unit_start_indicator ? "yes" : "no") << endl
	 << "transport priority: " << (transport_priority ? "yes" : "no") << endl
	 << "PID:                " << PID << endl
	 << "scrambling control: ";
    if (transport_scrambling_control==0) ostr << "no scrambling\n";
    else ostr << ((int)transport_scrambling_control) << endl;
    ostr << "adaptation field:   " << (has_adaptation_field ? "yes" : "no") << endl
	 << "continuity counter: " << ((int)continuity_counter) << endl
	 << "no. payload bytes:  " << payload_length << endl;
  }





  MP2_Stream_Type DetermineStreamType(InputStream& str)
  {
    uint8 buf[4];
    int size;
    int leading_zeros = 0;  // number of '0' bytes at beginning of file
    MP2_Stream_Type type;

    for (;;)
      {
	size = str.FillBuffer(buf,4);
	if (size<4)
	  {
	    type = MPEG_Unknown_Stream;
	    break;
	  }

	if (!leading_zeros && buf[0] == 0xFF && (buf[1] & 0xF0) == 0xF0)
	  {
	    type = MPEG_Audio_Stream;
	    break;
	  }

	if (!leading_zeros && buf[0] == 0x47)
	  {
	    type = MPEG_Transport_Stream;
	    break;
	  }

	if (buf[0] == 0 && buf[1] == 0 && buf[2] == 0)
	  {
	    leading_zeros++;
	    str.Pushback(&buf[1],3);
	    continue;
	  }

	if (buf[0] == 0 && buf[1] == 0 && buf[2] == 1)
	  {
	    if (buf[3] == MP2_StartCode_PackStart)
	      {
		type = MPEG_Program_Stream;
		break;
	      }

	    if (buf[3] == MP2_StartCode_SequenceHeader)
	      {
		type = MPEG_Video_Stream;
		break;
	      }
	  }

	type = MPEG_Unknown_Stream;
	goto type_found;
      }

  type_found:
    str.Pushback(buf,size);

    uint8 zero=0;
    for (int i=0;i<leading_zeros;i++)
      str.Pushback(&zero,1);

    return type;
  }


  void   SkipToNextStartCode(InputStream& str)
  {
    uint8 buf[3];

    int size = str.FillBuffer(buf,3,3);
    if (size<3)
      return;

    for (;;)
      {
	if (buf[0]==0 && buf[1]==0 && buf[2]==1)
	  break;

	buf[0]=buf[1];
	buf[1]=buf[2];

	if (str.FillBuffer(&buf[2],1,1) != 1)
	  return;
      }

    str.Pushback(buf,3);
  }


  void   SkipZerosToNextStartCode(InputStream& str)
  {
    uint8 buf[3];

    int size = str.FillBuffer(buf,3,3);
    if (size<3)
      return;

    for (;;)
      {
	if (!(buf[0]==0 && buf[1]==0 && buf[2]==0))
	  break;

	if (str.FillBuffer(&buf[2],1,1) != 1)
	  return;
      }

    str.Pushback(buf,3);
  }



  /* state transition graph for start-code acception
     (see doc/startcode_automate.obj)
  */

  static struct
  {
    int next_0, next_1, next_other,   n_stay_in_buffer;
    bool accept;
  } state_table[] =
    {
      {  1,  1,  1,     0, false },
      /* {  0,  1,  1,     0, false },  * using this transition instead results in
	 leading zeros before start-codes */
      {  2,  1,  1,     0, false },
      {  3,  1,  1,     1, false },
      {  3,  4,  1,     2, false },
      { -1, -1, -1,     3, true  }
    };

#define READ_BLOCK_LENGTH 200

  void ReadUntilNextStartCode(ByteBuffer& bb, InputStream& str)
  {
    uint8 buf[READ_BLOCK_LENGTH];
    int   size = 0;
    int   pos  = 0;
    int   bufstart = 0;

    int state = 0;

    while (!str.IsEOF())
      {
	size = str.FillBuffer(&buf[bufstart], READ_BLOCK_LENGTH-bufstart);
	size+=bufstart;

	while (pos<size)
	  {
	    switch(buf[pos])
	      {
	      case 0:  state = state_table[state].next_0; break;
	      case 1:  state = state_table[state].next_1; break;
	      default: state = state_table[state].next_other; break;
	      }

	    if (state_table[state].accept)
	      {
		// next start-code found

		bb.AppendBytes(buf,pos-2);
		str.Pushback(&buf[pos-2],size-(pos-2));
		return;
	      }

	    pos++;
	  }

	// copy buffer to output except the bytes that have to stay in the buffer

	int n_stay = state_table[state].n_stay_in_buffer;
	bb.AppendBytes(buf,size - n_stay);
	for (int i=0;i<n_stay;i++)
	  buf[i] = buf[size-n_stay+i];

	pos      = n_stay;
	bufstart = n_stay;
	size     = n_stay;
      }

    bb.AppendBytes(buf,size); // append all remaining bytes
  }


  void   TruncateUntilStartCode(ByteBuffer& bb)
  {
    int     len = bb.AskLength();
    uint8* data = bb.AskData();

    int start = 0;
    bool found = false;

    while (start <= len-4)
      {
	if (data[start+0] == 0 &&
	    data[start+1] == 0 &&
	    data[start+2] == 1)
	  {
	    found=true;
	    break;
	  }

	start++;
      }

    if (start==0)
      return;

    if (!found) start=len;

    for (int i=0;i<len-start;i++)
      data[i]=data[i+start];

    bb.TruncateBuffer(start);
  }

}
