/********************************************************************************
  mpeg2/syntax/system.hh
    MPEG system stream constants.

  purpose:
    Definition of system stream constants, system header structures,
    and some utility functions for stream parsing.

  notes:

  to do:

  author(s):
   - Dirk Farin, dirk.farin@gmx.de

  modifications:
   03/Mar/2002 - Dirk Farin
     - new code
 ********************************************************************************
    Copyright (C) 2002  Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef LIBVIDEOCODING_MPEG2_SYNTAX_SYSTEM_HH
#define LIBVIDEOCODING_MPEG2_SYNTAX_SYSTEM_HH

#include <libvideogfx/types.hh>
#include <libvideogfx/error.hh>
#include <libvideogfx/utility/bitstream/bytebuffer.hh>
#include <libvideocoding/fileio/inputstream.hh>

#include <iostream>
#include <iomanip>

namespace videocoding {
  using namespace videogfx;


  // -------------- startcodes ----------------

  enum MP2_StartCode
    {
      // video start-codes

      MP2_StartCode_PictureStart   = 0x00,
      MP2_StartCode_UserData       = 0xB2,
      MP2_StartCode_SequenceHeader = 0xB3,
      MP2_StartCode_SequenceError  = 0xB4,
      MP2_StartCode_ExtensionStart = 0xB5,
      MP2_StartCode_SequenceEnd    = 0xB7,
      MP2_StartCode_GroupStart     = 0xB8,

      // system start-codes

      MP2_StartCode_ISO11172_End      = 0xB9,
      MP2_StartCode_PackStart         = 0xBA,
      MP2_StartCode_SystemHeaderStart = 0xBB,

      // packet start-codes

      MP2_StartCode_Video0  = 0xE0,
      MP2_StartCode_Video1  = 0xE1,
      MP2_StartCode_Video2  = 0xE2,
      MP2_StartCode_Audio0  = 0xC0,
      MP2_StartCode_Audio1  = 0xC1,
      MP2_StartCode_Audio2  = 0xC2,

      MP2_StartCode_Program_Stream_Map = 0xBC,
      MP2_StartCode_Private_Stream_1   = 0xBD,
      MP2_StartCode_Padding_Stream     = 0xBE,
      MP2_StartCode_Private_Stream_2   = 0xBF,

      MP2_StartCode_ECM_Stream       = 0xF0,
      MP2_StartCode_EMM_Stream       = 0xF1,
      MP2_StartCode_DSMCC_Stream     = 0xF2,
      MP2_StartCode_IEC13522_Stream  = 0xF3,
      MP2_StartCode_H222_Type_A      = 0xF4,
      MP2_StartCode_H222_Type_B      = 0xF5,
      MP2_StartCode_H222_Type_C      = 0xF6,
      MP2_StartCode_H222_Type_D      = 0xF7,
      MP2_StartCode_H222_Type_E      = 0xF8,
      MP2_StartCode_Ancillary_Stream = 0xF9,
      MP2_StartCode_Program_Stream_Directory = 0xFF
    };

  inline bool IsMPEG_SLICE_START (MP2_StartCode code) { return code>=0x01 && code<=0xAF; }
  inline bool IsMPEG_VIDEO_STREAM(MP2_StartCode code) { return code>=0xE0 && code<=0xEF; }
  inline bool IsMPEG_AUDIO_STREAM(MP2_StartCode code) { return code>=0xC0 && code<=0xDF; }
  inline bool IsMPEG_PACKETCODE  (MP2_StartCode code) { return code>=0xBC && code<=0xFF; }


  // ---------- extension_start_code_identifier -----------

  enum MP2_ExtensionID
    {
      MP2_ExtID_SequenceExtension         = 0x1,
      MP2_ExtID_SequenceDisplayExtension  = 0x2,
      MP2_ExtID_QuantMatrixExtension      = 0x3,
      MP2_ExtID_CopyrightExtension        = 0x4,
      MP2_ExtID_SequenceScalableExtension = 0x5,
      MP2_ExtID_PictureDisplayExtension   = 0x7,
      MP2_ExtID_PictureCodingExtension    = 0x8,
      MP2_ExtID_PictureSpatialScalableExtension  = 0x9,
      MP2_ExtID_PictureTemporalScalableExtension = 0xA
    };


  // ---------- trick mode control -----------

  enum MP2_TrickModeControl
    {
      MP2_TrickMode_FastForward = 0x0,
      MP2_TrickMode_SlowMotion  = 0x1,
      MP2_TrickMode_FreezeFrame = 0x2,
      MP2_TrickMode_FastReverse = 0x3,
      MP2_TrickMode_SlowReverse = 0x4
    };


  // ---------- field control -----------

  enum MP2_FieldControl
    {
      MP2_FieldControl_TopFieldOnly    = 0x0,
      MP2_FieldControl_BottomFieldOnly = 0x1,
      MP2_FieldControl_CompleteFrame   = 0x2,
    };


  // ---------- coefficient selection -----------

  enum MP2_CoefficientSelection
    {
      MP2_CoefficientSelection_OnlyDC     = 0x0,
      MP2_CoefficientSelection_FirstThree = 0x1,
      MP2_CoefficientSelection_FirstSix   = 0x2,
      MP2_CoefficientSelection_All        = 0x3
    };



  typedef int64 PTS;
  typedef int64 DTS;


  struct MP2_SystemPacket
  {
    MP2_SystemPacket() : input_filepos(-1) { }
    virtual ~MP2_SystemPacket() { }

    virtual void Show(std::ostream& ostr) const { ostr << "(no header information implemented)\n"; }

    int64 input_filepos;   // File position where this system packet started. -1 = unknown

  protected:
    // static ByteBufferPool s_bytebufferpool;   // WARNING: don't use this; static members don't get initialized because of gcc bug
  };



  // ------------- program stream ----------------

  struct MP2_PackHeader : public MP2_SystemPacket
  {
    MP2_PackHeader() : n_stuffing_bytes(0) { }

    int64 SCR;
    int32 mux_rate;
    int   n_stuffing_bytes;

    void  Read(InputStream&);   // read pack header from input stream
    void  Show(std::ostream&) const;
  };


  struct MP2_SystemHeader : public MP2_SystemPacket
  {
    MP2_SystemHeader();

    int   rate_bound;
    int8  audio_bound;
    int8  video_bound;
    bool  fixed_bitrate;
    bool  CSPS_flag;
    bool  audio_lock;
    bool  video_lock;
    bool  packet_rate_restriction_flag;
	  
    int32 STD_buffer_size_bound[0x100-0xBC]; /* "stream id" - 0xBC / unit: bytes */

    void  Read(InputStream&);   // read from input stream
    void  Show(std::ostream&) const;
  };


  struct MP2_Iso11172EndCode : public MP2_SystemPacket
  {
    void  Show(std::ostream& ostr) const { ostr << "ISO-11172 end code\n"; }
  };


  struct MP2_PES_Packet : public MP2_SystemPacket
  {
    MP2_PES_Packet() { /* data.AttachToPool(&s_bytebufferpool); */ }

    uint8  stream_id;

    uint2  PES_scrambling_control;
    bool   PES_priority;
    bool   data_alignment_indicator;
    bool   copyright;
    bool   original;

    bool   has_PTS;
    uint64 PTS;

    bool   has_DTS;
    uint64 DTS;

    bool   has_ESCR;
    uint64 ESCR;

    bool   has_ES_rate;
    uint32 ES_rate;

    bool                     has_DSM_trick_mode;
    MP2_TrickModeControl     trick_mode_control;
    // fast forward/reverse
    MP2_FieldControl         field_id;    // + freeze frame
    bool                     intra_slice_refresh;
    MP2_CoefficientSelection frequency_truncation;
    // slow forward/reverse
    uint5                    rep_cntrl;

    bool   has_additional_copy_info;
    uint7  additional_copy_info;

    bool   has_PES_CRC;
    uint16 previous_PES_packet_CRC;

    bool   has_PES_extension;
    bool   has_PES_private_data;
    uint8  PES_private_data[16];
    bool   has_pack_header_field;
    // TODO: pack header
    bool   has_program_packet_sequence_counter;
    uint7  program_packet_sequence_counter;
    bool   MPEG1_MPEG2_identifier;
    uint6  original_stuff_length;
    bool   has_PSTD_buffer;
    uint32 STDBufferSize;
    bool   has_PES_extension_flag_2;

    int32 data_start_offset; // number of bytes between 'input_filepos' and this data

    ByteBuffer data; /* The actual packet data. Note that this can be larger
			than it is allowed by the standard, as the system
			decoder is free to create pseudo system packets that
			do not underlie any limitations. */

    void  Reset();
    void  Read(InputStream&);   // read from input stream
    void  Show(std::ostream&) const;
  };


  inline std::ostream& operator<<(std::ostream& ostr,const MP2_SystemPacket& pck)
  {
    ostr << "pos: " << std::hex << pck.input_filepos << std::dec << std::endl;
    pck.Show(ostr);
    return ostr;
  }


  // ------------- transport stream ----------------

  struct MP2_Transport_Packet : public MP2_SystemPacket
  {
    bool   transport_error_indicator;
    bool   payload_unit_start_indicator;
    bool   transport_priority;
    uint13 PID;
    uint2  transport_scrambling_control;
    bool   has_payload;
    bool   has_adaptation_field;
    uint4  continuity_counter;

    uint8  payload[184];
    int    payload_length;

    void  Read(InputStream&);   // read from input stream
    void  Show(std::ostream&) const;
  };


  // ------------- misc functions ------------------

  enum MP2_Stream_Type { MPEG_Program_Stream, MPEG_Transport_Stream,
			 MPEG_Video_Stream,   MPEG_Audio_Stream,
			 MPEG_Unknown_Stream };

  MP2_Stream_Type DetermineStreamType(InputStream&);
  void   SkipToNextStartCode(InputStream&);      // skip all kind of data until next start-code
  void   SkipZerosToNextStartCode(InputStream&); // skip all-zero bytes until start-code or junk is found
  void   ReadUntilNextStartCode(ByteBuffer&, InputStream&);
  void   TruncateUntilStartCode(ByteBuffer&);

}

#endif
